// $Id: SampleEventHandler.C,v 1.10 2006/05/11 22:39:37 emarkus Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

// A sample client for ORTS that shows how to access client data

// todo/questions:
//
// game scripts:
// - what to do with workers during build time
// - similarly, how to define mining?
// - how is attack range defined (distance between centers/hulls)?
//
// client:
// - how to access global player data?
// - how to compute the set of visible tiles?
// - how to compute tiles visible from a unit?
// - how to decide what can be attacked by a specific unit?

// FORS/FORALL/ERR macros are defined in orts3/libs/kernel/src/Global.H

#include "SampleEventHandler.H"
#include "GameStateModule.H"
//#include "GfxModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"

using namespace std;

// tile types from core.bp
// fixme: game script reflection would help here
// in the competition we won't use water tiles
// TILE_CLIFF: non-traversable by ground units

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

class MyMap
{
public:
	int map[32][32]; // containing "densities" or "weights" or "costs" (-1 = inaccessible)
					// yes, hardcoding is lame

	void init(Map<GameTile> *);

	void add(int x, int y, int w)
	{
		if(x<0 || x>31 || y<0 || y>31) return;

		if(map[x][y]!=-1)
			map[x][y]+=w;
	}
	void add(int x, int y, int w, int r)
	{
		--r;
		for(int j=y-r;j<=y+r;j++)
			for(int i=x-r;i<=x+r;i++)
				add(i,j,w/((i-x)*(i-x)+(j-y)*(j-y)+1));
	}
};

void MyMap::init(Map<GameTile> *_map)
{
	for(int i=0;i<32;i++)
		for(int j=0;j<32;j++)
		{
			GameTile &tile=_map->get_tile(i,j);
			if(tile.get_max_h()>0)
				map[i][j]=-1;
			else
				map[i][j]=1;
		}

	for(int y=0;y<32;y++)
	{
		printf("y=%2d ",y);
		for(int x=0;x<32;x++)
			if(map[x][y]==-1)
				printf("0");
			else
				printf("1");
		printf("\n");
	}
}

struct QueueElem
{
	int cost; // fixme: should be float
	ScalarPoint p;
	QueueElem *prev; // on the path
	int depth;

	int index; // isn't really needed?

	QueueElem(){};
	QueueElem(int _cost, ScalarPoint _p, QueueElem* _prev, int _depth, int _index)
		:cost(_cost),p(_p),prev(_prev), depth(_depth),index(_index) {}
};

class QueueElemCompare
{
public:
	bool operator()(QueueElem* const& pnode1, QueueElem* const& pnode2)
	{
		const QueueElem &node1=*pnode1, &node2=*pnode2;
		int c1=node1.cost;
		int c2=node2.cost;
		if(c1!=c2)
			return (c1<c2);

//		return &node1<&node2;
//		return (node1.p.y*32+node1.p.x)<(node2.p.y*32+node2.p.x);
		return node1.index<node2.index;
	}
};

// statistics
	int finds;
	int depths;
	int looks;
	
class MyPathFind
{
public:
	MyMap *oldmap;
	MyMap *map;

	int elem_index;
	
	typedef std::multiset<QueueElem*, QueueElemCompare> tqueue;
	typedef std::set<int> tvisited;

	tqueue *open;
//	std::hash_map<int> *visited;
	tvisited *visited;

	MyPathFind()
	{
		open=NULL; visited=NULL;
		finds=depths=looks=0;
	}


	void init(Map<GameTile> *m) 
	{
		oldmap=new MyMap();
		oldmap->init(m);
		
		map=new MyMap();
		map->init(m);
	}

	void apply_gamestate(Game *game) 
		// fixme: run it
	{
//		cout << "pre-Apllying gamestate" << endl;
		
		map=new MyMap(*oldmap);

		int n=game->get_player_num();
	
//		cout << "Apllying gamestate" << endl;
		
		FORS(i,(n+1))
			FORALL(game->get_objs(i),it)
			{
				int w,r;

				r=w=0;
				GameObj *gob=(*it)->get_GameObj();
				if(gob->bp_name()=="worker")
				{
					w=2; r=2;
//					cout << "worker" << endl;
				}
				else if(gob->bp_name()=="sheep")
				{
					w=3; r=1;
//					cout << "sheep" << endl;
				}
				else if(gob->bp_name()=="mineral")
				{
					w=4; r=2;
//					cout << "mineral" << endl;
				}

				if(r>0)
				{
					int x=*gob->sod.x;
					int y=*gob->sod.y;
					xy2t(x,y);
					map->add(x,y,w,r);
				}
			}

//		cout << "Gamestate applied" << endl;
	}

	bool find_path(GameObj *start, GameObj *end, Game *state, ScalarPoint *path, int steps);
	void add2queue(int x, int y, QueueElem *prev, int w);

	void xy2t(int &x, int &y)
	{
		x+=8;
		y+=8;
//		x=x+8; y=y+8;
        x=x/16; y=y/16; // what did I say about hardcoding?
		if(x>31) x=31;
		if(y>31) y=31;
	}

	void t2xy(int &x, int &y)
	{
		x=x*16+8;
		y=y*16+8;
	}

	int t2h(int x, int y)
	{
		return y*32+x;
	}
};

bool MyPathFind::find_path(GameObj *start, GameObj *end, Game *, ScalarPoint *path, int steps)
{
    int sx=*start->sod.x, sy=*start->sod.y;
	xy2t(sx,sy);

	int ex=*end->sod.x, ey=*end->sod.y;
	xy2t(ex,ey);

	elem_index=0;
	
	if(open!=NULL) delete open;
	if(visited!=NULL) delete visited;

	open=new tqueue();
	visited=new tvisited();

	QueueElem qe_start(0,ScalarPoint(sx,sy),NULL,0,-1);
	open->insert(&qe_start);


	QueueElem *next;

	while(1)
	{
//		unsigned int os=open->size();

		if(open->empty())
			return false;
		tqueue::iterator inext=open->begin();
        next=*inext; // I hope I know what I'm doing ;)
		open->erase(inext);
	
//		assert(open->size()!=os);
		
		int px=next->p.x, py=next->p.y;
		if(px==ex && py==ey)
			break; //we found it :)
	
//		char text[1000];
//		sprintf(text,"Visited: %d, open %d, visiting: %dx%d(%d)",visited->size(),open->size(),px,py,next->index);
//		cout << text << endl;
		
		if(visited->find(t2h(px,py))!=visited->end()) // we found better way, so throw this away
			continue;
		
		visited->insert(t2h(px,py)); // mark as visited

		// add all neighbors
//		for(int y=py-1; y<=py+1;y++)
//			for(int x=px-1; x<=px+1; x++)
//				if(x!=px||y!=py)
//	                add2queue(x,y,next);
		add2queue(px-1,py-1,next,3);	               
		add2queue(px+1,py-1,next,3);	               
		add2queue(px-1,py+1,next,3);	               
		add2queue(px+1,py+1,next,3);
		add2queue(px-1,py,next,2);
		add2queue(px+1,py,next,2);
		add2queue(px,py-1,next,2);
		add2queue(px,py+2,next,2);

/*		tqueue::iterator iter=open->begin();
		cout << "In queue: ";
		for(;iter!=open->end();iter++)
		{
			QueueElem *e=*iter;
			printf("%d = %dx%d, ",e->index,e->p.x,e->p.y);
//			printf("%d = %dx%d, ",e->index,elems[e->p.x,e->p.y);
		}
		cout << endl;*/
	}

//	cout << "OK! I'm free at last!" << endl;
	
// we have to fix coordinates (tiles->(x,y))
	next->p.x=*end->sod.x;
	next->p.y=*end->sod.y;

	finds++;
	looks+=elem_index;
	depths+=next->depth;
	
	while(1)
	{
//		cout << "Current depth " << next->depth << ", index: " << next->index << endl;
        if(next->depth < steps)
			path[next->depth]=next->p;
		assert(next!=next->prev);
		next=next->prev;
		if(next==NULL)
			break;
		else
			t2xy(next->p.x,next->p.y);
	}
	return true;
}

void MyPathFind::add2queue(int x, int y, QueueElem *prev,int w)
{
//	QueueElem *elems=new QueueElem[32*32*8];
	static QueueElem elems[32*32*8];
 
//	cout << "trying to add "<<x<<"x"<<y<<": ";
	
	if(x<0 || x>31) return;
	if(y<0 || y>31) return;

	if(map->map[x][y]==-1)
	{
//		cout << "cant go there" << endl;
		return; // inaccessible
	}

	if(visited->find(t2h(x,y))!=visited->end())
	{
//		cout << "been here" << endl;
		return; // we've seen it before
	}

	int new_cost=prev->cost+map->map[x][y]*w;

	assert(elem_index<32*32*8);
	elems[elem_index]=QueueElem(new_cost,ScalarPoint(x,y),prev,prev->depth+1,elem_index);

	assert(&elems[elem_index]!=prev);
	
//	unsigned int os=open->size();
	open->insert(&elems[elem_index]); // FIXME: it should really check if it's not a duplicate...
//	assert(os!=open->size());

//	char text[1000];
//	sprintf(text,"added %dx%d(%d)",elems[elem_index].p.x,elems[elem_index].p.y,elems[elem_index].index);
//	cout << text << endl;
	
	elem_index++;
}

MyPathFind *pathfinder; // this is sooooo lame... there's just *NO* reason to not make it SampleEventHandler's field

void SampleEventHandler::init_pathfinder()
{
	pathfinder=new MyPathFind();
	pathfinder->init((Map<GameTile>*)&state.gsm->get_game().get_map());
}

bool SampleEventHandler::handle_event(const Event& e)
{
	//	cout << "This is *creed's* event-handler" << endl;
	if (e.get_who() == GameStateModule::FROM) {

		//cout << "Event: " << e.who() << " " << e.what() << endl;

		if (e.get_what() == GameStateModule::STOP_MSG) {
			state.quit = true; return true;
			//prints statistics
/*			cout << "Pathfinder statistics: " << endl;
			cout << "Used " << pathfinder->finds << " times." << endl;
			cout << "Average depth of solution: " << pathfinder->depths/pathfinder->finds << endl;
			cout << "Average seen tiles: " << pathfinder->looks/pathfinder->finds << endl;*/
		}

		if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
			state.quit = state.error = true; return true;
			//prints statistics
/*			cout << "Pathfinder statistics: " << endl;
			cout << "Used " << pathfinder->finds << " times." << endl;
			cout << "Average depth of solution: " << pathfinder->depths/pathfinder->finds << endl;
			cout << "Average seen tiles: " << pathfinder->looks/pathfinder->finds << endl;*/
		}

		if (e.get_what() == GameStateModule::VIEW_MSG) {

			compute_actions();
			state.gsm->send_actions();

			Game &game = state.gsm->get_game();	

			// gfx and frame statistics

			//      if (state.gfxm) state.gfxm->process_changes();

			sint4 vf = game.get_view_frame();
			sint4 af = game.get_action_frame();
			sint4 sa = game.get_skipped_actions();
			sint4 ma = game.get_merged_actions();

			if ((vf % 20) == 0) cout << "[frame " << vf << "]" << endl;

			if (vf != af || sa || ma > 1) {

				cout << "frame " << vf;
				if (af < 0) cout << " [no action]";
				else if (vf != af) cout << " [behind by " << (vf - af) << " frame(s)]";
				if (sa) cout << " [skipped " << sa << "]";
				if (ma > 1) cout << " [merged " << ma << "]";
				cout << endl;
			}

			/*      if (state.gfxm) {

			// don't draw if we are far behind
			if (abs(vf-af) < 10) {
			state.gfxm->draw();
			state.just_drew = true;
			}
			}*/

			return true;
		}
	}

	return false;
}

void SampleEventHandler::init_miner(GameObj* miner)
{
	if(assignments.find(miner)==assignments.end())
	{
		Job *job=new Job();
		job->target=NULL;
		job->laststopped=true; // to prevent first stop-check
		job->job=JOB_GOMINE;
		assignments[miner]=job;
	}
}

void SampleEventHandler::get_target(int &tx, int &ty, GameObj* miner)
{
	static set<GameObj*>::iterator i=resources.begin();

	bool wasistopped=false;
// initialize if isn't
	init_miner(miner);
	Job *job=assignments[miner];
// check if I was stopped
	if((*miner->sod.x!=job->gx || *miner->sod.y!=job->gy) && job->laststopped==false)
	{
		wasistopped=true;
		job->laststopped=true;
		cout << "Oh man! I was stopped!";

		char text[1000];
		sprintf(text,"(%dx%d instead of %dx%d)",*miner->sod.x,*miner->sod.y,job->gx,job->gy);
		cout << text << endl;
	}

	
	if(wasistopped)
	{
		ServerObjData &sod=miner->sod;
		int rx,ry; 
		real8 rdist=1000000000000000000000.; // big real ;)
		GameObj *res=NULL;
		// find nearest resource
		for (set<GameObj*>::iterator i = resources.begin(); i != resources.end();++i)
		{
			ServerObjData &sod2=(*i)->sod;

			real8 d2=(*sod.x-*sod2.x)*(*sod.x-*sod2.x)+
				(*sod.y-*sod2.y)*(*sod.y-*sod2.y);

			if(d2<rdist)
			{
				rdist=d2;
				rx=*sod2.x;
				ry=*sod2.y;
				res=*i;
			}
		}

		if(job->job==JOB_GOMINE && Game::distance(*miner,*res)<=2) // if we're looking for minerals and they're close enough
		{
			cout << "Jeee! I hit the minerals :) Gimme, gimme!" << endl;
			job->target=res; // little trick ;)
			return;
		}
		else
		{ // just pick any goal at random
#define LEAP_X 20
#define LEAP_Y 20
			tx=*miner->sod.x + (rand.rand_uint4()%(2*LEAP_X))-LEAP_X;
			ty=*miner->sod.y + (rand.rand_uint4()%(2*LEAP_Y))-LEAP_Y;
			return;
		}
	}
	
	if(job->target==NULL)
	{
		if(job->job==JOB_GOMINE)
		{
	        job->target=*i;
			i++; if(i==resources.end()) i=resources.begin();
		}
		else if(job->job==JOB_GOHOME)
		{
			job->target=base;			
		}
		else
		{
			cout << "I don't know what to do!!!";
			return;
		}
	}

	// ok, we finally got to the state where we know where we want to go :) let's find the way!

	//GameObj* mineral=job->target; // obsolete
	//tx=*mineral->sod.x;
	//ty=*mineral->sod.y;

	ScalarPoint path[5];
//	cout << "Starting pathfinding!" << endl;
	if(pathfinder->find_path(miner,job->target,NULL,path,5)==false)
	{
		cout << "Help! I can't get there!!! Choosing another target!" << endl;
		
		if(job->job==JOB_GOMINE)
		{
			job->target=*i;
			i++; if(i==resources.end()) i=resources.begin();
		}
		
		tx=*miner->sod.x;
		ty=*miner->sod.y;
	}
	else
	{
//		cout << "Path found!" << endl;
        tx=path[2].x; job->gx=tx;
		ty=path[2].y; job->gy=ty;
	}

	job->laststopped=false;
	
	return;
}

//real8 SampleEventHandler::distance2(GameObj* gob1, GameObj* gob2) // obsolete ;)
//{
//	ServerObjData &sod1=gob1->sod;
//	ServerObjData &sod2=gob2->sod;
//
//	real8 dx=*sod1.x-*sod2.x;
//	real8 dy=*sod1.y-*sod2.y;
//	return dx*dx+dy*dy;
//}
//
void SampleEventHandler::compute_actions()
{
//	cout << endl;

//	cout << "compute actions" << endl;

	const GameChanges &changes = state.gsm->get_changes();

	// my code

	FORALL(changes.new_objs, i)
	{
		GameObj* gob=(*i)->get_GameObj();
		if(gob->bp_name() == "mineral")
			resources.insert(gob);
		else if(gob->bp_name() == "controlCenter")
			base=gob;
	}

	FORALL(changes.dead_objs, i)
	{
		GameObj *gob=(*i)->get_GameObj();
		if(gob->bp_name()=="mineral")
			resources.erase(gob);
		cout << "removed mineral" << endl;
	}

	Game &game = state.gsm->get_game();
	const sint4 cid = game.get_client_player();
	const Game::ObjCont &objs = game.get_objs(cid);

	if(pathfinder==NULL)
		init_pathfinder();
	pathfinder->apply_gamestate(&game);
	
	FORALL (objs, it) {
		GameObj *gob = (*it)->get_GameObj();
		if(*gob->sod.speed==0 && gob->sod.in_game && !gob->is_dead() && gob->bp_name()=="worker")
		{
			ScalarPoint t;

			init_miner(gob);
			GameObj *target=assignments[gob]->target;
			
			switch(assignments[gob]->job)
			{
				case JOB_MINE:
					if(gob->component("pickaxe")->get_int("active")!=0)
					{
//						cout << "I'm still mining!" << endl;
					}
					else
					{
						cout << "Finished mining!" << endl;
						cout << "I have " << gob->get_int("minerals") << "minerals" << endl;
						assignments[gob]->job=JOB_GOHOME;
					}
					break;

				case JOB_GOMINE:
					if(target!=NULL && Game::distance(*target,*gob)<=2)
					{
						cout << "I'm near minerals! mining!" << endl;

						Vector<sint4> args;
						args.push_back(game.get_cplayer_info().get_id(target));
						gob->component("pickaxe")->set_action("mine",args);

						assignments[gob]->job=JOB_MINE;
						assignments[gob]->target=NULL;
					}
					else
					{
						get_target(t.x,t.y,gob);

						Vector<sint4> args;
						args.push_back(t.x);
						args.push_back(t.y);
						args.push_back(*gob->sod.max_speed);
	
						gob->set_action("move", args);
					}
					break;

				case JOB_GOHOME:
					if(target!=NULL && Game::distance(*target,*gob)<=3)
					{
						cout << "I'm near base! returning!" << endl;

						Vector<sint4> args;
						args.push_back(game.get_cplayer_info().get_id(target));
						gob->set_action("return_resources",args);						
		
						assignments[gob]->job=JOB_RETURN;
						assignments[gob]->target=NULL;
					}
					else
					{
						get_target(t.x,t.y,gob);

                        Vector<sint4> args;
                        args.push_back(t.x);
                        args.push_back(t.y);
                        args.push_back(*gob->sod.max_speed);

                        gob->set_action("move", args);
					}
					break;

				case JOB_RETURN:
					if(gob->get_int("minerals")==0)
					{
						cout << "Returing finished" << endl;

						assignments[gob]->job=JOB_GOMINE;
					}
					else
					{
						cout << "I'm still returning!" << endl;
					}
					break;

				default: cout << "I don't know what to do!!!" << endl;
			}
//			break; // just one!
		}
//		if(gob->bp_name()=="worker") break; // comment it and let the fun begin!
	}


	//FORALL (objs, it) {
	//	GameObj *gob = (*it)->get_GameObj();
	//	ServerObjData &sod=gob->sod;

	//	int rx,ry; 
	//	real8 rdist=1000000000000000000000.; // big real ;)
	//	// find nearest resource
	//	//	FORALL(resources,res)
	//	for (set<GameObj*>::iterator i = resources.begin(); i != resources.end();++i)
	//	{
	//		ServerObjData &sod2=(*i)->sod;

	//		real8 d2=(*sod.x-*sod2.x)*(*sod.x-*sod2.x)+
	//			(*sod.y-*sod2.y)*(*sod.y-*sod2.y);

	//		if(d2<rdist)
	//		{
	//			rdist=d2;
	//			rx=*sod2.x;
	//			ry=*sod2.y;
	//		}
	//	}
	//	t.x=rx; t.y=ry;

	//	Vector<sint4> params;
	//	params.push_back(t.x);
	//	params.push_back(t.y);
	//	params.push_back(*gob->sod.max_speed);

	//	gob->set_action("move", params);
	//}
	return;
	// ends here

	bool cinfo;
	Options::get("-cinfo", cinfo);

	if (cinfo) {

		//	  cout << "creed's cinfo";
		// - what changed in the last simulation tick

		if (!changes.new_tile_indexes.empty())
			cout << "#new tiles = " << changes.new_tile_indexes.size() << endl;

		if (!changes.new_boundaries.empty())
			cout << "#new boundaries = " << changes.new_boundaries.size() << endl;

		// fixme: what about removed tiles and removed boundaries?
		// currently tiles and boundaries are never removed

		if (!changes.new_objs.empty())
		{
			cout << "#new objects = " << changes.new_objs.size() << endl;

			FORALL(changes.new_objs, i)
			{
				GameObj* gob=(*i)->get_GameObj();
				ServerObjData &sod=gob->sod;

				if(gob->bp_name()=="mineral")
				{
					cout << "new mineral at " << *sod.x << "x" << *sod.y << endl;
				}		  
			}	  
		}

		if (!changes.changed_objs.empty())
			cout << "#changed objects = " << changes.changed_objs.size() << endl;

		if (!changes.vanished_objs.empty())
			cout << "#vanished objects = " << changes.vanished_objs.size() << endl;

		if (!changes.dead_objs.empty())
			cout << "#dead objects = " << changes.dead_objs.size() << endl;

	}

//	const Game &game = state.gsm->get_game();
//	const sint4 cid = game.get_client_player();

	// ***************** terrain ******************

	{
		const Map<GameTile> &map = game.get_map();

		bool minfo;
		Options::get("-minfo", minfo);

		if (minfo) {

			// - map dimensions

			cout << "Map width: = " << map.get_width()  << " tiles" << endl;
			cout << "Map height: " << map.get_height() << " tiles" << endl;  

			// height field scale: HEIGHT_MULT = one tile width up, constant for now
			cout << "height mult: " << GameConst::HEIGHT_MULT << endl; 

			// - new/updated tile indexes

			cout << "# of new/updated tiles " << changes.new_tile_indexes.size() << endl;
			cout << "Index(es): ";
			FORALL (changes.new_tile_indexes, it) cout << *it << " ";
			cout << endl;

			// - new boundaries

			const Vector<ScriptObj*> &boundaries = changes.new_boundaries;

			FORALL (boundaries, it) {

				const GameObj *g = (*it)->get_GameObj();
				if (g == 0) continue;

				cout << "new boundary (" << *g->sod.x1 << "," << *g->sod.y1 << ")-("
					<< *g->sod.x2 << "," << *g->sod.y2 << ")"
					<< endl;
			}

			// - all boundaries

			const Game::ObjCont &all_boundaries = game.get_boundaries();
			cout << "total # of boundaries = " << all_boundaries.size() << endl;
		}


		bool tinfo;
		Options::get("-tinfo", tinfo);

		if (tinfo) {

			// - all tiles

			FORS (y, map.get_height()) {
				FORS (x, map.get_width()) {
					const GameTile &tile = map(x,y);
					cout << "[" << x << "," << y << "]: ";

					sint4 hNW, hNE, hSE, hSW;
					Tile::Split split;

					// get corner heights and split type
					tile.get_topo(hNW, hNE, hSE, hSW, split);

					// tile types: see enum on top
					// split types: see orts3/libs/kernel/src/Tile.H
					// if split == Tile::NO_SPLIT => typeW is the tile type

					cout << "split:" << split 
						<< " type w:" << tile.get_typeW()
						<< " type e:" << tile.get_typeE()
						<< " heights:" << hNW << " " << hNE << " " << hSE << " " << hSW
						<< endl;
				}

				cout << endl;
			}

			cout << endl;
		}


		// - currently visible tiles

		// ... to be implemented
	}

	// ***************** objects ******************

	bool oinfo;
	Options::get("-oinfo", oinfo);

	if (oinfo) {

		// units under client player control

		const Game::ObjCont &objs = game.get_objs(cid);

		FORALL (objs, it) {

			const GameObj *gob = (*it)->get_GameObj();
			if (!gob) ERR("not a game object!?");

			const ServerObjData &sod = gob->sod;

			// - fundamental properties

			// blueprint name ~ type name
			cout << "bp: " << gob->bp_name();

			if (!sod.in_game) { cout << " no gameobj" << endl; continue; }

			if (*sod.x < 0) { cout << "not on playfield" << endl; continue; }

			// object on playfield

			// sod.shape and sod.zcat values defined in orts3/libs/kernel/src/Object.H

			cout << "zcat: " << *sod.zcat
				<< " max-speed: " << *sod.max_speed
				<< " speed: " << *sod.speed 
				<< " moving: " << *sod.is_moving 
				<< " sight: " << *sod.sight;

			if (*sod.shape == Object::CIRCLE) {
				cout << " CIRCLE loc: (" << *sod.x << "," << *sod.y << ")"
					<< " r: "<< *sod.radius;

			} else if (*sod.shape == Object::RECTANGLE) {
				cout << " RECT dim: (" << *sod.x1 << "," << *sod.y1 << ")-("
					<< *sod.x2 << "," << *sod.y2 << ")";
			} else {
				cout << " UNKNOWN SHAPE";
			}

			// - other object properties

			// heading: 0 = stopped, 1...GAME_CONST::HEADING_N
			const sintptr *hd = gob->get_const_int_ptr("heading");
			if (hd) cout << " hd: " << GameConst::angle_from_dir(*hd, GameConst::HEADING_N);

			// ... more examples

			cout << endl;
		}
	}

	// - what an object currently sees

	// ... to be implemented

	// - setting object actions

	const Map<GameTile> &map = game.get_map();
//	const Game::ObjCont &objs = game.get_objs(cid);
	const sint4 points_x = map.get_width()  * game.get_tile_points();
	const sint4 points_y = map.get_height() * game.get_tile_points();  

	FORALL (objs, it) {

		GameObj *gob = (*it)->get_GameObj();
		ServerObjData &sod=gob->sod;

		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {

			ScriptObj *weapon = gob->component("weapon");
			if (!weapon) continue;

			const sint4 n = game.get_player_num();
			GameObj *target = 0;

			if (n > 1) { // there are other players

				// todo: aim at weakest opponent unit within range

				sintptr *prange = weapon->get_int_ptr("max_ground_range");
				if (!prange) continue;

				const sint4 range2 = *prange * *prange;

				FORS (i, n) {

					if (i == cid) continue;

					// check for targets (could be done much faster)

					const Game::ObjCont &opp_objs = game.get_objs(i);

					FORALL (opp_objs, it) {

						GameObj *gob2 = (*it)->get_GameObj();
						if (gob2 && gob2->sod.in_game && *gob2->sod.x >= 0 && !gob2->is_dead()) {
							real8 d2 =
								square(real8(*gob->sod.x - *gob2->sod.x)) +
								square(real8(*gob->sod.y - *gob2->sod.y));

							if (d2 <= range2) {
								// target within range
								target = gob2;
							}
						}
					}
				}
			}

			if (target) {

				cout << "attack" << endl;

				// attack target

				// fixme: observe cooldown

				//cout << "launch " << sqrt(min_d) << " " << sqrt(attack_d) << endl;
				Vector<sint4> args;
				args.push_back(game.get_cplayer_info().get_id(target));
				gob->component("weapon")->set_action("attack", args);
				continue;
			}

			if (*gob->sod.speed == 0) {

				// object not dead and speed == 0 => move action

				// todo: attack weakest opponent in range

				ScalarPoint t;

				if ((rand.rand_uint4() % 100) < 33) {

					// random target on fringe (x or y unchanged)

					switch (rand.rand_uint4() & 3) {
	  case 0: t.x = *gob->sod.x; t.y = 0; break;
	  case 1: t.x = *gob->sod.x; t.y = points_y-1; break;
	  case 2: t.x = 0; t.y = *gob->sod.y; break;           
	  case 3: t.x = points_x-1; t.y = *gob->sod.y; break;
					}

				} else {

					// random target on fringe

					switch (rand.rand_uint4() & 3) {

	  case 0:
		  t.x = 0;
		  t.y = rand.rand_uint4() % points_y;
		  break;

	  case 1:
		  t.x = points_x - 1;
		  t.y = rand.rand_uint4() % points_y;
		  break;

	  case 2:
		  t.x = rand.rand_uint4() % points_x;
		  t.y = 0;
		  break;

	  case 3:
		  t.x = rand.rand_uint4() % points_x;
		  t.y = points_y - 1;
		  break;
					}
				}

				int rx,ry; 
				real8 rdist=1000000000000000000000.; // big real ;)
				// find nearest resource
				//	FORALL(resources,res)
				for (set<GameObj*>::iterator i = resources.begin(); i != resources.end();++i)
				{
					ServerObjData &sod2=(*i)->sod;

					real8 d2=(*sod.x-*sod2.x)*(*sod.x-*sod2.x)+
						(*sod.y-*sod2.y)*(*sod.y-*sod2.y);

					if(d2<rdist)
					{
						rdist=d2;
						rx=*sod2.x;
						ry=*sod2.y;
					}
				}
				t.x=rx; t.y=ry;

				Vector<sint4> params;
				params.push_back(t.x);
				params.push_back(t.y);
				params.push_back(*gob->sod.max_speed);

				gob->set_action("move", params);
			}
		}
	}

	// - global player/game variables

	// ... to be implemented
}
